import dataclasses
import logging
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union

import datasets
from pie_core import Annotation, AnnotationLayer, Document, annotation_field
from pie_modules.annotations import BinaryRelation, LabeledMultiSpan, LabeledSpan
from pie_modules.documents import TextBasedDocument

from pie_datasets import GeneratorBasedBuilder

logger = logging.getLogger(__name__)


@dataclasses.dataclass(eq=True, frozen=True)
class BratAttribute(Annotation):
    annotation: Annotation
    label: str
    value: Optional[str] = None
    score: Optional[float] = dataclasses.field(default=None, compare=False)

    def resolve(self) -> Any:
        value = self.value if self.value is not None else True
        return value, self.label, self.annotation.resolve()


@dataclasses.dataclass(eq=True, frozen=True)
class BratNote(Annotation):
    annotation: Annotation
    label: str
    value: str

    def resolve(self) -> Any:
        return self.value, self.label, self.annotation.resolve()


@dataclasses.dataclass
class BratDocument(TextBasedDocument):
    spans: AnnotationLayer[LabeledMultiSpan] = annotation_field(target="text")
    relations: AnnotationLayer[BinaryRelation] = annotation_field(target="spans")
    span_attributes: AnnotationLayer[BratAttribute] = annotation_field(target="spans")
    relation_attributes: AnnotationLayer[BratAttribute] = annotation_field(target="relations")
    notes: AnnotationLayer[BratNote] = annotation_field(
        targets=["spans", "relations", "span_attributes", "relation_attributes"]
    )


@dataclasses.dataclass
class BratDocumentWithMergedSpans(TextBasedDocument):
    spans: AnnotationLayer[LabeledSpan] = annotation_field(target="text")
    relations: AnnotationLayer[BinaryRelation] = annotation_field(target="spans")
    span_attributes: AnnotationLayer[BratAttribute] = annotation_field(target="spans")
    relation_attributes: AnnotationLayer[BratAttribute] = annotation_field(target="relations")
    notes: AnnotationLayer[BratNote] = annotation_field(
        targets=["spans", "relations", "span_attributes", "relation_attributes"]
    )


def dl2ld(dict_of_lists: Dict[str, List[Any]]) -> List[Dict[str, Any]]:
    return [dict(zip(dict_of_lists, t)) for t in zip(*dict_of_lists.values())]


def ld2dl(
    list_fo_dicts: List[Dict[str, Any]], keys: Optional[List[str]] = None
) -> Dict[str, List[Any]]:
    keys = keys or list(list_fo_dicts[0])
    return {k: [dic[k] for dic in list_fo_dicts] for k in keys}


def example_to_document(
    example: Dict[str, Any], merge_fragmented_spans: bool = False
) -> Union[BratDocument, BratDocumentWithMergedSpans]:
    if merge_fragmented_spans:
        doc = BratDocumentWithMergedSpans(text=example["context"], id=example["file_name"])
    else:
        doc = BratDocument(text=example["context"], id=example["file_name"])

    spans: Dict[str, LabeledSpan] = dict()
    span_locations: List[Tuple[Tuple[int, int], ...]] = []
    span_texts: List[str] = []
    for span_dict in dl2ld(example["spans"]):
        starts: List[int] = span_dict["locations"]["start"]
        ends: List[int] = span_dict["locations"]["end"]
        slices = tuple(zip(starts, ends))
        span_locations.append(slices)
        span_texts.append(span_dict["text"])
        # sanity check
        span_text_parts = [doc.text[start:end] for start, end in slices]
        joined_span_texts_stripped = " ".join(span_text_parts).strip()
        span_text_stripped = span_dict["text"].strip()
        if joined_span_texts_stripped != span_text_stripped:
            logger.warning(
                f"joined span parts do not match stripped span text field content. "
                f'joined_span_texts_stripped: "{joined_span_texts_stripped}" != stripped "text": "{span_text_stripped}"'
            )
        if merge_fragmented_spans:
            if len(starts) > 1:
                # check if the text in between the fragments holds only space
                merged_content_texts = [
                    doc.text[start:end] for start, end in zip(ends[:-1], starts[1:])
                ]
                merged_content_texts_not_empty = [
                    text.strip() for text in merged_content_texts if text.strip() != ""
                ]
                if len(merged_content_texts_not_empty) > 0:
                    logger.warning(
                        f"document '{doc.id}' contains a non-contiguous span with text content in between "
                        f"(will be merged into a single span): "
                        f"newly covered text parts: {merged_content_texts_not_empty}, "
                        f"merged span text: '{doc.text[starts[0]:ends[-1]]}', "
                        f"annotation: {span_dict}"
                    )
            # just take everything
            start = min(starts)
            end = max(ends)
            span = LabeledSpan(start=start, end=end, label=span_dict["type"])
        else:
            span = LabeledMultiSpan(slices=slices, label=span_dict["type"])
        spans[span_dict["id"]] = span

    doc.spans.extend(spans.values())
    doc.metadata["span_ids"] = list(spans.keys())
    doc.metadata["span_locations"] = span_locations
    doc.metadata["span_texts"] = span_texts

    relations: Dict[str, BinaryRelation] = dict()
    for rel_dict in dl2ld(example["relations"]):
        arguments = dict(zip(rel_dict["arguments"]["type"], rel_dict["arguments"]["target"]))
        assert set(arguments) == {"Arg1", "Arg2"}
        head = spans[arguments["Arg1"]]
        tail = spans[arguments["Arg2"]]
        rel = BinaryRelation(head=head, tail=tail, label=rel_dict["type"])
        relations[rel_dict["id"]] = rel

    doc.relations.extend(relations.values())
    doc.metadata["relation_ids"] = list(relations.keys())

    equivalence_relations = dl2ld(example["equivalence_relations"])
    if len(equivalence_relations) > 0:
        raise NotImplementedError("converting equivalence_relations is not yet implemented")

    events = dl2ld(example["events"])
    if len(events) > 0:
        raise NotImplementedError("converting events is not yet implemented")

    attribute_annotations: Dict[str, Dict[str, BratAttribute]] = defaultdict(dict)
    attribute_ids = []
    for attribute_dict in dl2ld(example["attributions"]):
        target_id = attribute_dict["target"]
        if target_id in spans:
            target_layer_name = "spans"
            annotation = spans[target_id]
        elif target_id in relations:
            target_layer_name = "relations"
            annotation = relations[target_id]
        else:
            raise Exception("only span and relation attributes are supported yet")
        attribute = BratAttribute(
            annotation=annotation,
            label=attribute_dict["type"],
            value=attribute_dict["value"],
        )
        attribute_annotations[target_layer_name][attribute_dict["id"]] = attribute
        attribute_ids.append((target_layer_name, attribute_dict["id"]))

    doc.span_attributes.extend(attribute_annotations["spans"].values())
    doc.relation_attributes.extend(attribute_annotations["relations"].values())
    doc.metadata["attribute_ids"] = attribute_ids

    normalizations = dl2ld(example["normalizations"])
    if len(normalizations) > 0:
        raise NotImplementedError("converting normalizations is not yet implemented")

    id2annotation = {
        **spans,
        **relations,
        **attribute_annotations["spans"],
        **attribute_annotations["relations"],
    }

    doc.metadata["note_ids"] = []
    notes_dicts: Dict[Annotation, Dict[str, Any]] = dict()
    for note_dict in dl2ld(example["notes"]):
        target_id = note_dict["target"]
        if target_id not in id2annotation:
            raise Exception(f"note target {target_id} not found in any of the target layers")
        annotation = id2annotation[target_id]
        note = BratNote(
            annotation=annotation,
            label=note_dict["type"],
            value=note_dict["note"],
        )
        doc.notes.append(note)
        doc.metadata["note_ids"].append(note_dict["id"])

        # check for duplicates: this works only because the note was already added to the document
        # (annotations attached to different documents are never equal)
        if note in notes_dicts:
            prev_note = notes_dicts[note]
            logger.warning(
                f"document {doc.id}: annotation exists twice: {prev_note['id']} and {note_dict['id']} "
                f"are identical"
            )
        notes_dicts[note] = note_dict

    return doc


def document_to_example(
    document: Union[BratDocument, BratDocumentWithMergedSpans]
) -> Dict[str, Any]:
    annotation2id: Dict[Annotation, str] = dict()
    example = {
        "context": document.text,
        "file_name": document.id,
    }
    span_dicts: Dict[Union[LabeledSpan, LabeledMultiSpan], Dict[str, Any]] = dict()
    assert len(document.metadata["span_locations"]) == len(document.spans)
    assert len(document.metadata["span_texts"]) == len(document.spans)
    assert len(document.metadata["span_ids"]) == len(document.spans)
    for i, span in enumerate(document.spans):
        span_id = document.metadata["span_ids"][i]
        annotation2id[span] = span_id
        locations = tuple((start, end) for start, end in document.metadata["span_locations"][i])
        if isinstance(span, LabeledSpan):
            assert locations[0][0] == span.start
            assert locations[-1][1] == span.end
        elif isinstance(span, LabeledMultiSpan):
            assert span.slices == locations
        else:
            raise TypeError(f"span has unknown type [{type(span)}]: {span}")

        starts, ends = zip(*locations)
        span_dict = {
            "id": document.metadata["span_ids"][i],
            "locations": {
                "start": list(starts),
                "end": list(ends),
            },
            "text": document.metadata["span_texts"][i],
            "type": span.label,
        }
        if span in span_dicts:
            prev_ann_dict = span_dicts[span]
            ann_dict = span_dict
            logger.warning(
                f"document {document.id}: annotation exists twice: {prev_ann_dict['id']} and {ann_dict['id']} "
                f"are identical"
            )
        span_dicts[span] = span_dict
    example["spans"] = ld2dl(list(span_dicts.values()), keys=["id", "type", "locations", "text"])

    relation_dicts: Dict[BinaryRelation, Dict[str, Any]] = dict()
    assert len(document.metadata["relation_ids"]) == len(document.relations)
    for i, rel in enumerate(document.relations):
        rel_id = document.metadata["relation_ids"][i]
        annotation2id[rel] = rel_id
        arg1_id = span_dicts[rel.head]["id"]
        arg2_id = span_dicts[rel.tail]["id"]
        relation_dict = {
            "id": rel_id,
            "type": rel.label,
            "arguments": {
                "type": ["Arg1", "Arg2"],
                "target": [arg1_id, arg2_id],
            },
        }
        if rel in relation_dicts:
            prev_ann_dict = relation_dicts[rel]
            ann_dict = relation_dict
            logger.warning(
                f"document {document.id}: annotation exists twice: {prev_ann_dict['id']} and {ann_dict['id']} "
                f"are identical"
            )
        relation_dicts[rel] = relation_dict

    example["relations"] = ld2dl(list(relation_dicts.values()), keys=["id", "type", "arguments"])

    example["equivalence_relations"] = ld2dl([], keys=["type", "targets"])
    example["events"] = ld2dl([], keys=["id", "type", "trigger", "arguments"])

    all_attribute_annotations = {
        "spans": document.span_attributes,
        "relations": document.relation_attributes,
    }
    attribute_dicts: Dict[Annotation, Dict[str, Any]] = dict()
    attribute_ids_per_target = defaultdict(list)
    for target_layer, attribute_id in document.metadata["attribute_ids"]:
        attribute_ids_per_target[target_layer].append(attribute_id)

    for target_layer, attribute_ids in attribute_ids_per_target.items():
        attribute_annotations = all_attribute_annotations[target_layer]
        assert len(attribute_ids) == len(attribute_annotations)
        for i, attribute_annotation in enumerate(attribute_annotations):
            attribute_id = attribute_ids_per_target[target_layer][i]
            annotation2id[attribute_annotation] = attribute_id
            target_id = annotation2id[attribute_annotation.annotation]
            attribute_dict = {
                "id": attribute_id,
                "type": attribute_annotation.label,
                "target": target_id,
                "value": attribute_annotation.value,
            }
            if attribute_annotation in attribute_dicts:
                prev_ann_dict = attribute_dicts[attribute_annotation]
                ann_dict = attribute_annotation
                logger.warning(
                    f"document {document.id}: annotation exists twice: {prev_ann_dict['id']} and {ann_dict['id']} "
                    f"are identical"
                )
            attribute_dicts[attribute_annotation] = attribute_dict

    example["attributions"] = ld2dl(
        list(attribute_dicts.values()), keys=["id", "type", "target", "value"]
    )
    example["normalizations"] = ld2dl(
        [], keys=["id", "type", "target", "resource_id", "entity_id"]
    )

    notes_dicts: Dict[Annotation, Dict[str, Any]] = dict()
    for i, note_annotation in enumerate(document.notes):
        note_id = document.metadata["note_ids"][i]
        annotation2id[note_annotation] = note_id
        target_id = annotation2id[note_annotation.annotation]
        note_dict = {
            "id": note_id,
            "type": note_annotation.label,
            "target": target_id,
            "note": note_annotation.value,
        }
        if note_annotation in notes_dicts:
            prev_ann_dict = notes_dicts[note_annotation]
            logger.warning(
                f"document {document.id}: annotation exists twice: {prev_ann_dict['id']} and {note_dict['id']} "
                f"are identical"
            )
        notes_dicts[note_annotation] = note_dict
    example["notes"] = ld2dl(list(notes_dicts.values()), keys=["id", "type", "target", "note"])

    return example


class BratConfig(datasets.BuilderConfig):
    """BuilderConfig for BratDatasetLoader."""

    def __init__(self, merge_fragmented_spans: bool = False, **kwargs):
        """BuilderConfig for DocRED.

        Args:
          **kwargs: keyword arguments forwarded to super.
        """
        super().__init__(**kwargs)
        self.merge_fragmented_spans = merge_fragmented_spans


class BratBuilder(GeneratorBasedBuilder):
    DOCUMENT_TYPES = {
        "default": BratDocument,
        "merge_fragmented_spans": BratDocumentWithMergedSpans,
    }

    DEFAULT_CONFIG_NAME = "default"
    BUILDER_CONFIGS = [
        BratConfig(name="default"),
        BratConfig(name="merge_fragmented_spans", merge_fragmented_spans=True),
    ]

    BASE_DATASET_PATH = "DFKI-SLT/brat"
    BASE_DATASET_REVISION = (
        None  # it is highly recommended to set this to a commit hash in any derived builder
    )

    def _generate_document(self, example, **kwargs):
        return example_to_document(
            example, merge_fragmented_spans=self.config.merge_fragmented_spans
        )

    def _generate_example(self, document: Document, **kwargs) -> Dict[str, Any]:
        if not isinstance(document, (BratDocument, BratDocumentWithMergedSpans)):
            raise TypeError(f"document type {type(document)} is not supported")
        return document_to_example(document)



from __future__ import annotations

import logging
from typing import Callable, Generic, TypeVar

from pie_core import Document
from pie_core.utils.hydra import resolve_target

logger = logging.getLogger(__name__)


D = TypeVar("D", bound=Document)


class Pipeline:
    def __init__(
        self,
        *processors_list: Callable[[Document], Document],
        **processors_dict: Callable[[Document], Document],
    ):
        if processors_list and processors_dict:
            raise ValueError("You cannot use both positional and keyword arguments.")
        if processors_list:
            self.processors = {
                f"processor_{i}": processor for i, processor in enumerate(processors_list)
            }
        else:
            self.processors = processors_dict

    def __call__(self, doc: Document) -> Document:
        for _, processor in self.processors.items():
            doc = processor(doc)
        return doc


class Caster(Generic[D]):
    def __init__(self, document_type: str | type[D], **kwargs):
        dt = resolve_target(document_type, full_key="")
        if not (isinstance(dt, type) and issubclass(dt, Document)):
            raise TypeError(
                f"(resolved) document_type must be a subclass of Document, but it is: {dt}"
            )
        self.document_type = dt
        self.kwargs = kwargs

    def __call__(self, doc: Document) -> D:
        return doc.as_type(new_type=self.document_type, **self.kwargs)


class Converter(Generic[D]):
    """Generic converter class that takes a function that converts a document to another document
    and passes all other arguments to the function.

    If the function argument is not a callable, it is assumed that it is a string and the function
    is resolved via hydra.
    """

    def __init__(self, function: str | Callable[[Document], D], **kwargs):
        if callable(function):
            self.convert = function
        else:
            self.convert = resolve_target(function, full_key="")
        self.kwargs = kwargs

    def __call__(self, doc: Document) -> D:
        return self.convert(doc, **self.kwargs)



import json
import logging
import os
import shutil
from abc import ABC, abstractmethod
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    SupportsIndex,
    Type,
    TypeVar,
    Union,
)

import datasets
from pie_core import (
    Document,
    EnterDatasetDictMixin,
    EnterDatasetMixin,
    ExitDatasetDictMixin,
    ExitDatasetMixin,
    WithDocumentTypeMixin,
)
from pie_core.utils.hydra import resolve_target, serialize_document_type

from .dataset import (
    Dataset,
    IterableDataset,
    concatenate_datasets,
    get_pie_dataset_type,
)

logger = logging.getLogger(__name__)

METADATA_FILE_NAME = "metadata.json"


D = TypeVar("D", bound=Document)


class DatasetDict(datasets.DatasetDict):
    def __getitem__(self, k) -> Union[Dataset, IterableDataset]:  # type: ignore
        """Returns an individual dataset split."""

        dataset = super().__getitem__(k)
        if isinstance(dataset, (Dataset, IterableDataset)):
            return dataset
        else:
            raise TypeError(f"dataset must be of type Dataset, but is {type(dataset)}")

    @classmethod
    def load_dataset(cls, *args, split=None, **kwargs) -> "DatasetDict":
        dataset_or_dataset_dict = datasets.load_dataset(*args, split=split, **kwargs)
        if isinstance(dataset_or_dataset_dict, (Dataset, IterableDataset)):
            if split is None:
                raise ValueError(
                    f"split must be provided if the loaded dataset is not a (Iterable)DatasetDict, "
                    f"but is {type(dataset_or_dataset_dict)}"
                )
            return cls({split: dataset_or_dataset_dict})
        elif isinstance(
            dataset_or_dataset_dict, (datasets.DatasetDict, datasets.IterableDatasetDict)
        ):
            for dataset in dataset_or_dataset_dict.values():
                if not isinstance(dataset, (Dataset, IterableDataset)):
                    raise TypeError(
                        f"expected pie_datasets.Dataset or pie_datasets.IterableDataset, but got {type(dataset)}"
                    )
            return cls(dataset_or_dataset_dict)
        else:
            raise TypeError(
                f"expected pie_datasets.DatasetDict, pie_datasets.IterableDatasetDict, pie_datasets.Dataset, "
                f"or pie_datasets.IterableDataset, but got {type(dataset_or_dataset_dict)}"
            )

    @classmethod
    def from_hf(
        cls,
        hf_dataset: Union[
            datasets.DatasetDict,
            datasets.IterableDatasetDict,
            Dict[str, datasets.Dataset],
            Dict[str, datasets.IterableDataset],
        ],
        document_type: Union[str, Type[Document]],
    ) -> "DatasetDict":
        """Creates a PIE DatasetDict from a HuggingFace DatasetDict, or IterableDatasetDict. If the
        input is a Dataset or IterableDataset, we create a DatasetDict with one split named
        "train".

        Args:
            hf_dataset: HuggingFace (Iterable)Dataset(Dict)
            document_type: document type of the dataset. Can be a subclass of Document or string that can be
                resolved to such a type.
        """

        doc_type = resolve_target(document_type)
        if not isinstance(doc_type, type) or not issubclass(doc_type, Document):
            raise TypeError(f"document_type must be a subclass of Document, but is {doc_type}")

        res = cls(
            {
                k: get_pie_dataset_type(v).from_hf_dataset(v, document_type=doc_type)
                for k, v in hf_dataset.items()
            }
        )
        return res

    @classmethod
    def from_json(  # type: ignore
        cls,
        document_type: Optional[Union[Type[Document], str]] = None,
        metadata_path: Optional[Union[str, Path]] = None,
        data_dir: Optional[str] = None,
        split: Optional[str] = None,
        **kwargs,
    ) -> "DatasetDict":
        """Creates a PIE DatasetDict from JSONLINE files. Uses `datasets.load_dataset("json")`
        under the hood. Requires a document type to be provided. If the document type is not
        provided, we try to load it from the metadata file.

        Args:
            document_type: document type of the dataset
            data_dir: Defining the `data_dir` of the dataset configuration. See datasets.load_dataset() for more
                information.
            metadata_path: path to the metadata file. Should point to a directory containing the metadata file
                `metadata.json`. Defaults to the value of the `data_dir` parameter.
            split: if provided, only the specified split is loaded. see `datasets.load_dataset()` for more information.
            **kwargs: additional keyword arguments for `datasets.load_dataset()`
        """

        # try to load metadata
        if metadata_path is None:
            metadata_path = data_dir
        if metadata_path is not None:
            metadata_file_name = Path(metadata_path) / METADATA_FILE_NAME
            if os.path.exists(metadata_file_name):
                with open(metadata_file_name) as f:
                    metadata = json.load(f)
                document_type = document_type or metadata.get("document_type", None)

        if document_type is None:
            raise ValueError(
                "document_type must be provided if it cannot be loaded from the metadata file"
            )

        hf_dataset = datasets.load_dataset("json", data_dir=data_dir, split=split, **kwargs)
        if isinstance(hf_dataset, (datasets.Dataset, datasets.IterableDataset)):
            if split is None:
                raise ValueError(
                    f"split must be provided if the loaded dataset is not a (Iterable)DatasetDict, "
                    f"but is {type(hf_dataset)}"
                )
            hf_dataset = {split: hf_dataset}
        return cls.from_hf(hf_dataset, document_type=document_type)

    def to_json(self, path: Union[str, Path], mode: str = "a", **kwargs) -> None:
        """Serializes the DatasetDict. We convert all documents with `.asdict()` and dump them with
        `json.dump()` to one JSONLINE file per split. If there is already serialized data in the
        output directory, we append the new data to the existing files.

        Args:
            path: path to the output directory
            mode: mode for writing the data. Can be either "a" (append) or "w" (overwrite). Default is "a".
            **kwargs: additional keyword arguments for `json.dump()`
        """

        path = Path(path)

        # save the metadata
        metadata = {"document_type": serialize_document_type(self.document_type)}

        if mode == "a":
            os.makedirs(path, exist_ok=True)
            if os.path.exists(path / METADATA_FILE_NAME):
                # load previous metadata
                with open(path / METADATA_FILE_NAME) as f:
                    previous_metadata = json.load(f)
                if previous_metadata != metadata:
                    raise ValueError(
                        f"The metadata file {path / METADATA_FILE_NAME} already exists, "
                        "but the content does not match the current metadata. Can not append "
                        "the current dataset to already serialized data."
                        f"\nprevious metadata: {previous_metadata}"
                        f"\ncurrent metadata: {metadata}"
                    )
            else:
                with open(path / METADATA_FILE_NAME, "w") as f:
                    json.dump(metadata, f, indent=2)
        elif mode == "w":
            # clear the output directory if it exists (via shutils)
            if os.path.exists(path):
                logger.warning(
                    f'Dataset serialization directory "{path}" already exists, removing it to overwrite existing files.'
                )
                shutil.rmtree(path)
            os.makedirs(path, exist_ok=True)

            with open(path / METADATA_FILE_NAME, "w") as f:
                json.dump(metadata, f, indent=2)
        else:
            raise ValueError(f'mode must be "a" (append) or "w" (overwrite), but is "{mode}".')

        # save the splits
        for split, dataset in self.items():
            split_path = path / split
            logger.info(f'serialize documents to "{split_path}" ...')
            os.makedirs(split_path, exist_ok=True)
            file_name = split_path / "documents.jsonl"
            write_mode = "a" if os.path.exists(file_name) else "w"
            with open(file_name, write_mode) as f:
                for doc in dataset:
                    f.write(json.dumps(doc.asdict(), **kwargs) + "\n")

    @property
    def document_type(self) -> Type[Document]:
        """Returns the document type of the dataset splits.

        Raises an error if there are no splits in the dataset or if the dataset splits have
        different document types.
        """

        if len(self) == 0:
            raise ValueError("dataset does not contain any splits, cannot determine document type")
        document_types = {ds.document_type for ds in self.values()}
        if len(document_types) > 1:
            raise ValueError(
                f"dataset contains splits with different document types: {document_types}"
            )
        return next(iter(document_types))

    @property
    def dataset_type(self) -> Union[Type[Dataset], Type[IterableDataset]]:
        """Returns the dataset type of the dataset splits, i.e. either `Dataset` or
        `IterableDataset`.

        Raises an error if there are no splits in the dataset or if the dataset splits have
        different dataset types.
        """

        if len(self) == 0:
            raise ValueError(
                "dataset does not contain any splits, cannot determine the dataset type"
            )
        dataset_types = {type(ds) for ds in self.values()}
        if len(dataset_types) > 1:
            raise ValueError(
                f"dataset contains splits with different dataset types: {dataset_types}"
            )
        return next(iter(dataset_types))

    def register_document_converter(
        self,
        converter: Union[Callable[..., D], Dict[str, str], str],
        document_type: Optional[Union[Type[D], str]] = None,
    ) -> "DatasetDict":
        """Register a converter function or field mapping for a target document type.

        Args:
            document_type: The target document type for which the converter should be registered. Can be a subclass
                of Document or string that can be resolved to such a type. If `None`, the document type is tried to be
                inferred from the converter function signature.
            converter: Either a function that converts a document of the document type of this dataset to a document
                of the target document_type, a string that can be resolved to such a function, or a field mapping
                (dict[str, str]) that maps fields of the document type of this dataset to fields of the target
                document_type.
        """
        resolved_document_type: Optional[Union[Type[D], Callable]] = None
        if document_type is not None:
            if isinstance(document_type, str):
                resolved_document_type = resolve_target(document_type)
            else:
                resolved_document_type = document_type
            if not (
                isinstance(resolved_document_type, type)
                and issubclass(resolved_document_type, Document)
            ):
                raise TypeError(
                    f"document_type must be or resolve to a subclass of Document, but is '{document_type}'"
                )

        resolved_converter: Union[Callable[..., Any], dict[str, str]]
        if isinstance(converter, str):
            resolved_converter = resolve_target(converter)
        else:
            resolved_converter = converter
        if not (callable(resolved_converter) or isinstance(resolved_converter, dict)):
            raise TypeError(
                f"converter must be a callable or a dict, but is {type(resolved_converter)}"
            )

        for ds in self.values():
            ds.register_document_converter(
                document_type=resolved_document_type, converter=resolved_converter
            )
        return self

    def to_document_type(
        self,
        document_type: Union[Type[Document], str, WithDocumentTypeMixin],
        downcast: bool = True,
        **kwargs,
    ) -> "DatasetDict":
        """Converts all documents in the dataset to a new document type using the best registered
        document converter.

        Args:
            document_type: document type to convert the documents to. Can be a subclass of Document
                or string that can be resolved to such a type, or a class that implements the
                `WithDocumentTypeMixin` interface.
            downcast: if `True` (default), the dataset is converted to the new document type even
                if it already has a document type that is a superclass of the new document type.
                If `False`, no conversion is done in this case.
        """

        if isinstance(document_type, WithDocumentTypeMixin):
            name = type(document_type).__name__
            if document_type.document_type is None:
                logger.warning(
                    f"{name} does not specify a document type. The dataset can "
                    f"not be automatically converted to a document type."
                )
                return self
            logger.info(
                f"convert the dataset to the document type that is specified by {name}: "
                f"{document_type.document_type}"
            )
            resolved_document_type = document_type.document_type
        elif isinstance(document_type, str):
            resolved_document_type = resolve_target(document_type)
        else:
            resolved_document_type = document_type
        if not (
            isinstance(resolved_document_type, type)
            and issubclass(resolved_document_type, Document)
        ):
            raise TypeError(
                f"document_type must be a document type or a string that can be resolved to such a type, "
                f"but got {document_type}."
            )

        if downcast:
            if resolved_document_type == self.document_type:
                logger.info(
                    f"The dataset has already the requested document type {document_type}."
                )
                return self
        else:
            if issubclass(resolved_document_type, self.document_type):
                logger.info(
                    f"The dataset has already the requested document type {document_type}."
                )
                return self

        result = type(self)(
            {
                name: ds.to_document_type(document_type=resolved_document_type, **kwargs)
                for name, ds in self.items()
            }
        )
        return result

    def map(  # type: ignore
        self,
        function: Optional[Union[Callable, str]] = None,
        result_document_type: Optional[Union[str, Type[Document]]] = None,
        set_batch_size_to_split_size: bool = False,
        **kwargs,
    ) -> "DatasetDict":
        """Applies a function to all documents in the dataset.

        If the function is an object and is derived from the following mixins, the respective logic
        is applied:
        - EnterDatasetMixin: `enter_dataset(dataset_split, split_name)` is called before the function is
            applied to a dataset split
        - ExitDatasetMixin: `exit_dataset(processed_dataset_split, split_name)` is called after the function
            is applied to a dataset split
        - EnterDatasetDictMixin: `enter_dataset_dict(dataset_dict)` is called before any dataset split is
            processed (and before any `enter_dataset()` is called)
        - ExitDatasetDictMixin: `exit_dataset_dict(processed_dataset_dict)` is called after all dataset splits
            are processed (and after all `exit_dataset()` are called)

        Args:
            function: function to apply to the documents. If `None`, the identity function is used. If `str`,
                the function is resolved from the global namespace.
            result_document_type: optional document type of the resulting dataset. Can be a subclass of Document or
                string that can be resolved to such a type. If not provided, it is tried to infer it from the
                function signature. If this is not possible, the document type of the input dataset
                is used.
            set_batch_size_to_split_size: If enabled, set the batch_size to the size of the respective split
                when calling map() on it. This is useful to transform whole splits when using it in
                combination with batched=True.
            **kwargs: additional keyword arguments for `datasets.Dataset.map()`
        """

        if function is not None:
            func = resolve_target(function)
            if not callable(func):
                raise TypeError(f"function must be callable, but is of type {type(func)}")
        else:

            def identity(x):
                # exclude from coverage because its usage happens in the map which is not collected
                return x  # pragma: no cover

            func = identity
        map_kwargs = dict(function=func, **kwargs)
        if result_document_type is not None:
            map_kwargs["result_document_type"] = resolve_target(result_document_type)

        if isinstance(func, EnterDatasetDictMixin):
            func.enter_dataset_dict(self)

        result_dict = {}
        for split, dataset in self.items():
            if isinstance(func, EnterDatasetMixin):
                func.enter_dataset(dataset=dataset, name=split)
            if set_batch_size_to_split_size:
                map_kwargs["batch_size"] = len(dataset)
            result_dict[split] = dataset.map(**map_kwargs)
            if isinstance(func, ExitDatasetMixin):
                func.exit_dataset(dataset=result_dict[split], name=split)

        result = type(self)(result_dict)

        if isinstance(func, ExitDatasetDictMixin):
            func.exit_dataset_dict(result)

        return result

    def select(
        self,
        split: str,
        start: Optional[SupportsIndex] = None,
        stop: Optional[SupportsIndex] = None,
        step: Optional[SupportsIndex] = None,
        **kwargs,
    ) -> "DatasetDict":
        """Reduce a certain dataset split to a selection of its documents. This is similar to the
        Huggingface `select()`, but adds optional parameters `start`, `stop`, `step` that will be
        used to create indices, if available.

        Args:
            split: name of the dataset split to modify
            start: optional start index of the selection
            stop: optional stop index of the selection
            step: optional step size of the selection
            **kwargs: additional keyword arguments for `datasets.Dataset.select()`
        """

        if stop is not None:
            range_args = [stop]
            if start is not None:
                range_args = [start] + range_args
            if step is not None:
                range_args = range_args + [step]
            kwargs["indices"] = range(*range_args)

        if "indices" in kwargs:
            result = type(self)(self)
            pie_split = result[split]
            if not isinstance(pie_split, Dataset):
                raise TypeError(
                    f"can only select from a Dataset, but the split '{split}' is of type {type(pie_split)}"
                )
            result[split] = Dataset.from_hf_dataset(
                dataset=pie_split.select(**kwargs),
                document_type=pie_split.document_type,
                document_converters=pie_split.document_converters,
            )
            return result
        else:
            if len(kwargs) > 0:
                logger.warning(
                    f"arguments for dataset.select() available, but they do not contain 'indices' which is required, "
                    f"so we do not call select. provided arguments: \n{json.dumps(kwargs, indent=2)}"
                )
            return self

    def rename_splits(
        self,
        mapping: Optional[Dict[str, str]] = None,
        keep_other_splits: bool = True,
    ) -> "DatasetDict":
        """Renames the dataset splits.

        Args:
            mapping: mapping from old split names to new split names.
            keep_other_splits: if `True` (default), splits not contained in `mapping` are kept in the dataset
        """

        if mapping is None:
            mapping = {}
        result = type(self)(
            {
                mapping.get(name, name): data
                for name, data in self.items()
                if name in mapping or keep_other_splits
            }
        )
        return result

    def add_test_split(
        self,
        source_split: str = "train",
        target_split: str = "test",
        **kwargs,
    ) -> "DatasetDict":
        """Adds a test split to the dataset by splitting the source split.

        Uses the Huggingface `train_test_split()` method.
        """

        pie_split = self[source_split]
        if not isinstance(pie_split, Dataset):
            raise TypeError(
                f"can only create a train-test-split from a Dataset, but the source split '{source_split}' is of type "
                f"{type(pie_split)}"
            )
        split_result_hf = pie_split.train_test_split(**kwargs)
        split_result = type(self)(
            {
                name: Dataset.from_hf_dataset(
                    ds,
                    document_type=pie_split.document_type,
                    document_converters=pie_split.document_converters,
                )
                for name, ds in split_result_hf.items()
            }
        )
        res = type(self)(self)
        res[source_split] = split_result["train"]
        res[target_split] = split_result["test"]
        split_sizes = {k: len(v) for k, v in res.items()}
        logger.info(f"dataset size after adding the split: {split_sizes}")
        return res

    def drop_splits(self, split_names: List[str]) -> "DatasetDict":
        """Drops splits from the dataset.

        Args:
            split_names: names of the splits to drop
        """

        result = type(self)({name: ds for name, ds in self.items() if name not in split_names})
        return result

    def concat_splits(self, splits: List[str], target: str) -> "DatasetDict":
        """Concatenates selected splits into a new split.

        Args:
            splits: names of the splits to concatenate
            target: name of the new split
        """

        if any(split not in self for split in splits):
            raise ValueError(
                f"not all splits to concatenate are present in the dataset: {splits}, {self.keys()}"
            )
        if len(splits) == 0:
            raise ValueError("please provide at least one split to concatenate")
        result = type(self)({name: ds for name, ds in self.items() if name not in splits})
        if not issubclass(self.dataset_type, Dataset):
            raise TypeError(
                f"can only concatenate splits if the dataset type is a Dataset, but it is {self.dataset_type}"
            )
        splits_to_concat: List[Dataset] = [self[name] for name in splits]  # type: ignore
        if any(self.dataset_type != type(ds) for ds in splits_to_concat):
            raise ValueError(
                f"not all splits to concatenate have the same dataset type: "
                f"{({name: type(self[name]) for name in splits})}"
            )
        document_converters = None
        for ds in splits_to_concat:
            if ds.document_converters is not None:
                if document_converters is None:
                    document_converters = {}
                document_converters.update(ds.document_converters)
        # TODO: why do we need to ignore the typing here?
        concatenated = datasets.concatenate_datasets(splits_to_concat)  # type: ignore
        if not issubclass(self.dataset_type, type(concatenated)):
            raise ValueError(
                f"concatenated dataset is not of the same type as the original dataset: "
                f"{self.dataset_type}, {type(concatenated)}"
            )
        result[target] = self.dataset_type.from_hf_dataset(
            concatenated, document_type=self.document_type, document_converters=document_converters
        )
        split_sizes = {k: len(v) for k, v in result.items()}
        logger.info(f"dataset size after concatenating splits: {split_sizes}")
        return result

    def filter(  # type: ignore
        self,
        split: str,
        function: Optional[Union[Callable[[Dict], bool], str]] = None,
        result_split_name: Optional[str] = None,
        **kwargs,
    ) -> "DatasetDict":
        """Filters a dataset split using a filter function.

        Note: In contrast to `map`, the filter function gets the example dict instead of a document as input
        because the PIE variant of `Dataset.filter()` is not yet implemented and, thus, the Huggingface
        variant is internally used instead.

        Args:
            split: name of the split to filter
            function: filter function that is called on each example dict. Can be provided as a callable or as a
                string that is resolved to a callable using `resolve_target()`.
            result_split_name: name of the split to store the filtered examples in. If `None`, the filtered examples
                are stored in the same split as the original examples.
        """

        if function is not None:
            # create a shallow copy to not modify the input
            result = type(self)(self)
            function = resolve_target(function)
            pie_split = result[split]
            # TODO: Implement pytorch_ie.Dataset.filter() in a similar way such as map() to make use of the
            #  document type. For now, the filter function is called directly on the HF dataset and thus needs to
            #  accept a dict as input.
            # we need to convert the dataset back to HF because the filter function internally uses map() which will
            # break if the PIE variant is used
            hf_split: Union[datasets.Dataset, datasets.IterableDataset]
            if isinstance(pie_split, Dataset):
                hf_split = datasets.Dataset(**Dataset.get_base_kwargs(pie_split))
            elif isinstance(pie_split, IterableDataset):
                hf_split = datasets.IterableDataset(**IterableDataset.get_base_kwargs(pie_split))
            else:
                raise ValueError(f"dataset split has unknown type: {type(pie_split)}")
            hf_split_filtered = hf_split.filter(function=function, **kwargs)
            target_split_name = result_split_name or split
            target_split = type(pie_split).from_hf_dataset(
                dataset=hf_split_filtered,  # type: ignore
                document_type=pie_split.document_type,
                document_converters=pie_split.document_converters,
            )
            # iterable datasets do not have a length
            if not isinstance(target_split, IterableDataset):
                logger.info(
                    f"filtered split [{target_split_name}] has {len(target_split)} entries"
                )
            result[target_split_name] = target_split
            return result
        else:
            return self

    def move_to_new_split(
        self,
        ids: Optional[List[str]] = None,
        filter_function: Optional[Union[Callable[[Dict[str, Any]], bool], str]] = None,
        source_split: str = "train",
        target_split: str = "test",
    ) -> "DatasetDict":
        """Moves examples from one split to another split. ids or a filter function can be provided
        to select the examples to move.

        Args:
            ids: list of ids of the examples to move
            filter_function: filter function that is called on each example dict. Can be provided as a callable or as a
                string that can be resolved to such a callable.
            source_split: name of the split to move the examples from
            target_split: name of the split to move the examples to
        """

        if filter_function is not None:
            filter_func = resolve_target(filter_function)
        else:
            if ids is None:
                raise ValueError("please provide either a list of ids or a filter function")

            ids_set = set(ids)

            def filter_with_ids(ex: Dict[str, Any]):
                # exclude from coverage because its usage happens in the map which is not collected
                return ex["id"] in ids_set  # pragma: no cover

            filter_func = filter_with_ids

        dataset_with_only_ids = self.filter(
            split=source_split,
            function=filter_func,
        )
        dataset_without_ids = self.filter(
            split=source_split,
            function=lambda ex: not filter_func(ex),
        )
        dataset_without_ids[target_split] = dataset_with_only_ids[source_split]

        split_sizes = {k: len(v) for k, v in dataset_without_ids.items()}
        logger.info(f"dataset size after moving to new split: {split_sizes}")
        return dataset_without_ids

    def move_shard_to_new_split(
        self,
        source_split: str = "train",
        target_split: str = "test",
        **shard_kwargs,
    ) -> "DatasetDict":
        """Shards the source split and moves one shard to the target split.

        Args:
            source_split: name of the split to move the shard from
            target_split: name of the new split to move the shard to
            shard_kwargs: additional keyword arguments for `datasets.Dataset.shard()`
        """
        pie_split = self[source_split]
        shard_result_hf = pie_split.shard(**shard_kwargs)
        shard_ids = [doc.id for doc in shard_result_hf]

        return self.move_to_new_split(
            ids=shard_ids, source_split=source_split, target_split=target_split
        )

    def cast_document_type(
        self, new_document_type: Union[Type[Document], str], **kwargs
    ) -> "DatasetDict":
        """Casts the document type of all splits to a new document type."""

        new_type = resolve_target(new_document_type)

        result = type(self)(
            {
                name: ds.cast_document_type(new_document_type=new_type, **kwargs)
                for name, ds in self.items()
            }
        )
        return result


def load_dataset(*args, **kwargs) -> Union[DatasetDict, Dataset, IterableDataset]:
    dataset_or_dataset_dict = datasets.load_dataset(*args, **kwargs)
    if isinstance(dataset_or_dataset_dict, (Dataset, IterableDataset)):
        return dataset_or_dataset_dict
    elif isinstance(dataset_or_dataset_dict, (datasets.DatasetDict, datasets.IterableDatasetDict)):
        for name, dataset in dataset_or_dataset_dict.items():
            if not isinstance(dataset, (Dataset, IterableDataset)):
                raise TypeError(
                    f'expected all splits to be {Dataset} or {IterableDataset}, but split "{name}" is of type '
                    f"{type(dataset)}"
                )
        return DatasetDict(dataset_or_dataset_dict)
    else:
        raise TypeError(
            f"expected datasets.load_dataset to return {datasets.DatasetDict}, {datasets.IterableDatasetDict}, "
            f"{Dataset}, or {IterableDataset}, but got {type(dataset_or_dataset_dict)}"
        )


def concatenate_dataset_dicts(
    inputs: Dict[str, DatasetDict],
    split_mappings: Dict[str, Dict[str, Union[str, List[str]]]],
    clear_metadata: bool,
):
    """Concatenate the splits of multiple dataset dicts into a single one. Dataset name will be
    saved in Metadata.

    Args:
        inputs: A mapping from dataset names to dataset dicts that contain the splits to concatenate.
        split_mappings: A mapping from target split name to mappings from input dataset name to
            source split name or list of names.
        clear_metadata: Whether to clear the metadata before concatenating.

    Returns: A dataset dict with keys in split_names as splits and content from the merged input
        dataset dicts.
    """

    input_splits: Dict[str, Dict[str, Union[Dataset, IterableDataset]]] = {}
    for target_split_name, mapping in split_mappings.items():
        input_splits[target_split_name] = {}
        for ds_name, source_split_name in mapping.items():
            if isinstance(source_split_name, str):
                input_splits[target_split_name][ds_name] = inputs[ds_name][source_split_name]
            elif isinstance(source_split_name, list):
                input_splits[target_split_name][ds_name] = concatenate_datasets(
                    [
                        inputs[ds_name][_source_split_name]
                        for _source_split_name in source_split_name
                    ],
                    clear_metadata=clear_metadata,
                )

    result = DatasetDict(
        {
            target_split_name: concatenate_datasets(dsets, clear_metadata=clear_metadata)
            for target_split_name, dsets in input_splits.items()
        }
    )

    return result



from typing import List

import pyarrow as pa
from datasets.formatting.formatting import Formatter
from pie_core import Document


class DocumentFormatter(Formatter[Document, list, List[Document]]):
    def __init__(self, document_type, features=None, **kwargs):
        super().__init__(features=None)
        self.document_type = document_type

    def format_row(self, pa_table: pa.Table) -> Document:
        row = self.python_arrow_extractor().extract_row(pa_table)
        return self.document_type.fromdict(row)

    def format_column(self, pa_table: pa.Table) -> list:
        return []

    def format_batch(self, pa_table: pa.Table) -> List[Document]:
        batch = self.simple_arrow_extractor().extract_batch(pa_table).to_pylist()
        return [self.document_type.fromdict(b) for b in batch]



import logging
from collections.abc import Sequence
from functools import wraps
from inspect import Signature, isclass, signature
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import datasets
from datasets.formatting import _register_formatter
from pie_core import Document
from pie_core.utils.dictionary import (
    dict_of_lists2list_of_dicts,
    list_of_dicts2dict_of_lists,
)

from .document_formatter import DocumentFormatter

logger = logging.getLogger(__name__)

_register_formatter(DocumentFormatter, "document")


def decorate_convert_document_back(f):
    """Decorate the mapped function, so that its return value is converted back to a dict.

    If the input is a list, it converts the list of Documents to a dict of lists.
    """

    @wraps(f)
    def decorated(item, *args, **kwargs):
        doc_or_docs = f(item, *args, **kwargs)
        if isinstance(doc_or_docs, Document):
            return doc_or_docs.asdict()
        elif isinstance(doc_or_docs, list):
            return list_of_dicts2dict_of_lists([e.asdict() for e in doc_or_docs])
        else:
            raise TypeError(
                f"The function {f} should return a Document or a list of Documents, but returned {type(doc_or_docs)}"
            )

    return decorated


def decorate_convert_to_document(f, document_type: Type[Document], batched: bool):
    """Decorate the mapped function, so that it converts a dict to a Document.

    If batched is True, it converts a list of dicts to a list of Documents.
    """

    @wraps(f)
    def decorated(item, *args, **kwargs):
        if batched:
            docs = [document_type.fromdict(e) for e in dict_of_lists2list_of_dicts(item)]
            return f(docs, *args, **kwargs)
        else:
            doc = document_type.fromdict(item)
            return f(doc, *args, **kwargs)

    return decorated


def decorate_convert_to_document_and_back(f, document_type: Type[Document], batched: bool):
    """Decorate the mapped function, so that it converts a dict to a Document and the result back
    to a dict.

    If batched is True, it converts a list of dicts to a list of Documents and the result back to a
    dict of lists.
    """

    @wraps(f)
    def decorated(item, *args, **kwargs):
        if batched:
            docs = [document_type.fromdict(e) for e in dict_of_lists2list_of_dicts(item)]
            mapped_docs = f(docs, *args, **kwargs)
            return list_of_dicts2dict_of_lists([e.asdict() for e in mapped_docs])
        else:
            doc = document_type.fromdict(item)
            mapped_doc = f(doc, *args, **kwargs)
            return mapped_doc.asdict()

    return decorated


def _check_fields_for_casting(
    field_mapping: Dict[str, str],
    current_document_type: Type[Document],
    new_document_type: Type[Document],
    column_names: list[str],
    check_annotation_types: bool = True,
) -> Tuple[Set[str], Set[str]]:
    original_fields = {field.name: field for field in current_document_type.fields()}
    new_fields = {field.name: field for field in new_document_type.fields()}
    hidden_fields = set(column_names) - set(original_fields)
    fields_to_map_not_in_original_fields = (
        set(field_mapping) - set(original_fields) - set(hidden_fields)
    )
    if len(fields_to_map_not_in_original_fields) > 0:
        raise ValueError(
            f"some fields to rename are not in the original document_type or hidden fields: "
            f"{fields_to_map_not_in_original_fields}"
        )
    mapped_but_not_in_new_fields = set(field_mapping.values()) - set(new_fields)
    if len(mapped_but_not_in_new_fields) > 0:
        raise ValueError(
            f"some renamed fields are not in the new document_type: {mapped_but_not_in_new_fields}"
        )
    original_fields_mapped = {
        field_mapping.get(f_name, f_name): f for f_name, f in original_fields.items()
    }
    added_field_names = set(new_fields) - set(original_fields_mapped)
    removed_field_names = set(original_fields) - set(new_fields) - set(field_mapping)

    # Sanity checks
    if check_annotation_types:
        kept_field_names = set(original_fields_mapped) & set(new_fields)
        for f_name_mapped in kept_field_names:
            f = original_fields_mapped[f_name_mapped]
            new_f = new_fields[f_name_mapped]
            if not (
                f.type == new_f.type
                and f.default == new_f.default
                and f.default_factory == new_f.default_factory
            ):
                raise ValueError(f"new field is not the same as old field:\n{new_f}\nvs\n{f}")

    return removed_field_names, added_field_names


def _infer_document_type_from_function_return(
    function: Callable, strict: bool = True
) -> Optional[Type[Document]]:
    # try to infer the document type from the return type annotation of function
    return_signature = signature(function).return_annotation
    if not return_signature == Signature.empty:
        if not isclass(return_signature) or not issubclass(return_signature, Document):
            msg = "the return type annotation of the function used with map is not a subclass of Document"
            if strict:
                raise TypeError(msg)
            else:
                logger.warning(msg)
                return None
        return return_signature
    return None


D = TypeVar("D", bound=Document)
DocumentConvertersType = Dict[Type[D], Union[Callable[..., D], Dict[str, str]]]


def _get_best_dataset_converter_with_types(
    dataset: Union["IterableDataset", "Dataset"],
    document_type: Union[Type[Document]],
) -> Tuple[Union[Callable[..., Document], Dict[str, str]], Type[Document], Type[Document]]:
    # first try to find an exact match
    if document_type in dataset.document_converters:
        return dataset.document_converters[document_type], document_type, document_type

    # then try to find a match with a superclass
    for registered_dt, candidate_converter in dataset.document_converters.items():
        if issubclass(registered_dt, document_type):
            return candidate_converter, document_type, registered_dt

    # then try to find a match with a subclass
    for registered_dt, candidate_converter in dataset.document_converters.items():
        if issubclass(document_type, registered_dt):
            return candidate_converter, document_type, registered_dt

    raise ValueError(
        f"No valid key (either subclass or superclass) was found for the document type '{document_type}' "
        f"in the document_converters of the dataset. Available keys: {set(dataset.document_converters)}. "
        f"Consider adding a respective converter to the dataset with "
        f"dataset.register_document_converter(my_converter_method) where my_converter_method should accept "
        f"{dataset.document_type} as input and return '{document_type}'."
    )


@overload
def dataset_to_document_type(
    dataset: "Dataset",
    document_type: Type[Document],
    **kwargs,
) -> "Dataset": ...


@overload
def dataset_to_document_type(
    dataset: "IterableDataset",
    document_type: Type[Document],
    **kwargs,
) -> "IterableDataset": ...


def dataset_to_document_type(
    dataset: Union["IterableDataset", "Dataset"],
    document_type: Type[Document],
    **kwargs,
) -> Union["IterableDataset", "Dataset"]:
    # do nothing if the document type is already the requested type
    if document_type == dataset.document_type:
        logger.info(f"The dataset has already the requested document type {document_type}.")
        return dataset

    converter, requested_type, registered_type = _get_best_dataset_converter_with_types(
        dataset=dataset,
        document_type=document_type,
    )

    result = dataset
    if callable(converter):
        result = result.map(
            function=converter,
            result_document_type=registered_type,
            fn_kwargs=kwargs,
        )
    else:
        result = result.cast_document_type(
            new_document_type=registered_type, field_mapping=converter, **kwargs
        )
    # if the type is not the same or a subclass of the requested type, try to cast (again)
    if not issubclass(registered_type, requested_type):
        result = result.cast_document_type(new_document_type=requested_type)

    # remove the document converters because they are not valid anymore
    result.document_converters = {}

    # remove features not declared in the target document type
    if result.features is not None:
        original_field_names = set(result.features)
        target_field_names = {field.name for field in document_type.fields()}
        remove_field_names = original_field_names - target_field_names
        result = result.remove_columns(list(remove_field_names))

    return result


def dataset_register_document_converter(
    dataset: Union["Dataset", "IterableDataset"],
    converter: Union[Callable[..., D], Dict[str, str]],
    document_type: Optional[Type[D]] = None,
) -> None:
    if callable(converter) and document_type is None:
        dt = _infer_document_type_from_function_return(converter)
    else:
        dt = document_type
    if not (isinstance(dt, type) and issubclass(dt, Document)):
        raise TypeError(
            f"the (inferred) document_type {dt} is not a subclass of Document. "
            "Please provide a document_type or a converter with a return type annotation."
        )
    dataset.document_converters[dt] = converter


class Dataset(datasets.Dataset, Sequence[D]):
    def __init__(
        self,
        document_type: Type[D],
        arrow_table: datasets.table.Table,
        info: Optional[datasets.DatasetInfo] = None,
        split: Optional[datasets.NamedSplit] = None,
        indices_table: Optional[datasets.table.Table] = None,
        fingerprint: Optional[str] = None,
        document_converters: Optional[DocumentConvertersType] = None,
    ):
        super().__init__(
            arrow_table=arrow_table,
            info=info,
            split=split,
            indices_table=indices_table,
            fingerprint=fingerprint,
        )

        self.document_type = document_type
        self.set_format("document", document_type=document_type)
        self.document_converters = document_converters or {}

    @classmethod
    def get_base_kwargs(cls, dataset: datasets.Dataset):
        return dict(
            arrow_table=dataset._data,
            info=dataset.info,
            split=dataset.split,
            indices_table=dataset._indices,
            fingerprint=dataset._fingerprint,
        )

    @classmethod
    def from_hf_dataset(
        cls,
        dataset: datasets.Dataset,
        document_type: Type[D],
        document_converters: Optional[DocumentConvertersType] = None,
    ) -> "Dataset":
        document_dataset = cls(
            document_type=document_type,
            document_converters=document_converters,
            **cls.get_base_kwargs(dataset),
        )
        return document_dataset

    @classmethod
    def from_documents(
        cls,
        documents: List[Document],
        document_converters: Optional[DocumentConvertersType] = None,
        **dataset_kwargs,
    ) -> "Dataset":
        """Create a Dataset from a list of documents. It wraps the Huggingface
        datasets.Dataset.from_list method, see the documentation for more details.

        Args:
            documents (List[Document]): A list of documents.
            document_converters (Optional[DocumentConvertersType], optional): A dictionary of document
                converters. Defaults to None.
            **dataset_kwargs: Additional arguments for the Huggingface dataset creation.

        Returns:
            Dataset: The created dataset.
        """

        if len(documents) == 0:
            raise ValueError("No documents to create dataset from")
        document_type = type(documents[0])
        data = [doc.asdict() for doc in documents]
        hf_dataset = datasets.Dataset.from_list(mapping=data, **dataset_kwargs)
        dataset = cls.from_hf_dataset(
            hf_dataset, document_type=document_type, document_converters=document_converters
        )
        return dataset

    def apply_hf_func(self, func, **kwargs) -> "Dataset":
        return Dataset.from_hf_dataset(
            func(self, **kwargs),
            document_type=self.document_type,
            document_converters=self.document_converters,
        )

    def register_document_converter(
        self,
        converter: Union[Callable[..., D], Dict[str, str]],
        document_type: Optional[Type[D]] = None,
    ) -> None:
        dataset_register_document_converter(
            dataset=self,
            converter=converter,
            document_type=document_type,
        )

    def to_document_type(
        self,
        document_type: Type[Document],
        **kwargs,
    ) -> "Dataset":
        return dataset_to_document_type(
            dataset=self,
            document_type=document_type,
            **kwargs,
        )

    def map_to_hf(
        self, function: Optional[Callable] = None, as_documents: bool = False, **map_kwargs
    ) -> datasets.Dataset:
        """Map the dataset using a function and return a Huggingface Dataset.

        Args:
            function (Optional[Callable], optional): The function to apply to the documents. Defaults to None.
            as_documents (bool, optional): Whether the function returns documents. Defaults to False.
            **map_kwargs: Additional keyword arguments for the Huggingface Dataset.map method.
        """
        if function is not None and as_documents:
            function = decorate_convert_document_back(function)
        return super().map(function=function, **map_kwargs)

    def map(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        with_rank: bool = False,
        input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        drop_last_batch: bool = False,
        remove_columns: Optional[Union[str, List[str]]] = None,
        keep_in_memory: bool = False,
        load_from_cache_file: Optional[bool] = None,
        cache_file_name: Optional[str] = None,
        writer_batch_size: Optional[int] = 1000,
        features: Optional[datasets.Features] = None,
        disable_nullable: bool = False,
        fn_kwargs: Optional[dict] = None,
        num_proc: Optional[int] = None,
        suffix_template: str = "_{rank:05d}_of_{num_proc:05d}",
        new_fingerprint: Optional[str] = None,
        desc: Optional[str] = None,
        as_documents: bool = True,
        result_document_type: Optional[Type[Document]] = None,
    ) -> "Dataset":
        dataset = self.map_to_hf(
            function=function,
            as_documents=as_documents,
            with_indices=with_indices,
            with_rank=with_rank,
            input_columns=input_columns,
            batched=batched,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            remove_columns=remove_columns,
            keep_in_memory=keep_in_memory,
            # ignore typing because typing in Huggingface Dataset.map() is incorrect
            load_from_cache_file=load_from_cache_file,  # type: ignore
            cache_file_name=cache_file_name,
            writer_batch_size=writer_batch_size,
            features=features,
            disable_nullable=disable_nullable,
            fn_kwargs=fn_kwargs,
            num_proc=num_proc,
            suffix_template=suffix_template,
            new_fingerprint=new_fingerprint,
            desc=desc,
        )

        if result_document_type is None:
            result_document_type = self.document_type

        return Dataset.from_hf_dataset(
            dataset,
            document_type=result_document_type,
            document_converters=self.document_converters,
        )

    def cast_document_type(
        self,
        new_document_type: Type[D],
        remove_columns: bool = False,
        field_mapping: Optional[Dict[str, str]] = None,
        keep_annotation_types: bool = False,
    ) -> "Dataset":
        field_mapping = field_mapping or {}

        removed_field_names, added_field_names = _check_fields_for_casting(
            field_mapping=field_mapping,
            current_document_type=self.document_type,
            new_document_type=new_document_type,
            column_names=self.column_names,
            check_annotation_types=keep_annotation_types,
        )

        new_hf_dataset = datasets.Dataset(**self.get_base_kwargs(self))

        if remove_columns:
            new_hf_dataset = new_hf_dataset.remove_columns(list(removed_field_names))

        rename_targets_already_in_columns = (
            set(field_mapping.values()) - set(field_mapping)
        ) & set(new_hf_dataset.column_names)
        if len(rename_targets_already_in_columns) > 0:
            raise ValueError(
                f"rename targets are already in column names: {rename_targets_already_in_columns}. Did you miss "
                f"to set remove_columns=True in a previous call of cast_document_type?"
            )

        new_hf_dataset = new_hf_dataset.rename_columns(field_mapping)
        for f_name in added_field_names:
            if f_name not in new_hf_dataset.column_names:
                # add empty columns
                new_hf_dataset = new_hf_dataset.add_column(
                    name=f_name, column=len(new_hf_dataset) * [{}]
                )
        new_dataset = Dataset.from_hf_dataset(
            new_hf_dataset,
            document_type=new_document_type,
            document_converters=self.document_converters,
        )

        return new_dataset


class IterableDataset(datasets.IterableDataset):
    def __init__(
        self,
        document_type: Type[Document],
        hidden_columns: Optional[Set[str]] = None,
        document_converters: Optional[DocumentConvertersType] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.document_type = document_type
        self._document_field_names = [field.name for field in document_type.fields()]
        self.hidden_columns = set()
        if hidden_columns is not None:
            self.hidden_columns.update(hidden_columns)
        self.document_converters = document_converters or {}

    @property
    def column_names(self) -> List[str]:
        return self._document_field_names + list(self.hidden_columns)

    @classmethod
    def get_base_kwargs(cls, dataset: datasets.IterableDataset):
        return dict(
            ex_iterable=dataset._ex_iterable,
            info=dataset.info,
            split=dataset.split,
            formatting=dataset._formatting,
            shuffling=dataset._shuffling,
            distributed=dataset._distributed,
            token_per_repo_id=dataset._token_per_repo_id,
        )

    @classmethod
    def from_hf_dataset(
        cls,
        dataset: datasets.IterableDataset,
        document_type: Type[Document],
        hidden_columns: Optional[Set[str]] = None,
        document_converters: Optional[DocumentConvertersType] = None,
    ) -> "IterableDataset":
        dataset = cls(
            document_type=document_type,
            hidden_columns=hidden_columns,
            document_converters=document_converters,
            **cls.get_base_kwargs(dataset),
        )
        return dataset

    @classmethod
    def from_documents(
        cls,
        documents: Callable,
        document_converters: Optional[DocumentConvertersType] = None,
        **dataset_kwargs,
    ) -> "IterableDataset":
        """Create an Iterable Dataset from a generator that yields documents. It wraps the
        Huggingface datasets.IterableDataset.from_generator method, see the documentation for more
        details.

        Args:
            documents (Callable): A generator function that `yields` documents.
            document_converters (Optional[DocumentConvertersType], optional): A dictionary of document
                converters. Defaults to None.
            **dataset_kwargs: Additional arguments for the Huggingface dataset creation.

        Returns:
            IterableDataset: The created iterable dataset.
        """

        # get first document to infer the document type
        try:
            gen_kwargs = dataset_kwargs.get("gen_kwargs", {})
            first_doc = next(documents(**gen_kwargs))
        except StopIteration:
            raise ValueError("No documents to create dataset from")
        document_type = type(first_doc)

        # wrap the generator to yield dictionaries
        def wrapped_documents_generator(**kwargs):
            for doc in documents(**kwargs):
                yield doc.asdict()

        hf_dataset = datasets.IterableDataset.from_generator(
            wrapped_documents_generator, **dataset_kwargs
        )
        dataset = cls.from_hf_dataset(
            hf_dataset,
            document_type=document_type,
            document_converters=document_converters,
        )
        return dataset

    def __iter__(self):
        for example in iter(super().__iter__()):
            yield self.document_type.fromdict(example)

    def register_document_converter(
        self,
        converter: Union[Callable[..., D], Dict[str, str]],
        document_type: Optional[Type[D]] = None,
    ) -> None:
        dataset_register_document_converter(
            dataset=self,
            converter=converter,
            document_type=document_type,
        )

    def to_document_type(
        self,
        document_type: Type[Document],
        **kwargs,
    ) -> "IterableDataset":
        return dataset_to_document_type(
            dataset=self,
            document_type=document_type,
            **kwargs,
        )

    def map_to_hf(
        self,
        function: Optional[Callable] = None,
        as_documents: bool = False,
        batched: bool = False,
        **map_kwargs,
    ) -> datasets.IterableDataset:
        """Map the dataset using a function and return a Huggingface IterableDataset.

        Args:
            function (Optional[Callable], optional): The function to apply to the documents. Defaults to None.
            as_documents (bool, optional): Whether the function returns documents. Defaults to False.
            batched (bool, optional): Whether to apply the function in batches. Defaults to False.
            **map_kwargs: Additional keyword arguments for the Huggingface IterableDataset.map method.
        """
        if function is not None:
            if as_documents:
                function = decorate_convert_to_document_and_back(
                    function, document_type=self.document_type, batched=batched
                )
            else:
                function = decorate_convert_to_document(
                    function, document_type=self.document_type, batched=batched
                )
        return super().map(function=function, batched=batched, **map_kwargs)

    def map(  # type: ignore
        self,
        function: Optional[Callable] = None,
        as_documents: bool = True,
        result_document_type: Optional[Type[Document]] = None,
        **map_kwargs,
    ) -> "IterableDataset":
        dataset_mapped = self.map_to_hf(function=function, as_documents=as_documents, **map_kwargs)

        if result_document_type is None:
            result_document_type = self.document_type

        return IterableDataset.from_hf_dataset(
            dataset_mapped,
            document_type=result_document_type,
            document_converters=self.document_converters,
        )

    def apply_hf_func(self, func, **kwargs) -> "IterableDataset":
        return IterableDataset.from_hf_dataset(
            func(self, **kwargs),
            document_type=self.document_type,
            hidden_columns=self.hidden_columns,
            document_converters=self.document_converters,
        )

    def cast_document_type(
        self,
        new_document_type: Type[D],
        remove_columns: bool = False,
        field_mapping: Optional[Dict[str, str]] = None,
        keep_annotation_types: bool = False,
    ) -> "IterableDataset":
        field_mapping = field_mapping or {}

        removed_field_names, added_field_names = _check_fields_for_casting(
            field_mapping=field_mapping,
            current_document_type=self.document_type,
            new_document_type=new_document_type,
            column_names=self.column_names,
            check_annotation_types=keep_annotation_types,
        )
        hidden_columns = set(self.hidden_columns)
        new_hf_dataset = datasets.IterableDataset(**self.get_base_kwargs(self))

        if remove_columns:
            new_hf_dataset = new_hf_dataset.remove_columns(column_names=list(removed_field_names))
        else:
            hidden_columns.update(removed_field_names)

        rename_targets_already_in_columns = (
            set(field_mapping.values()) - set(field_mapping)
        ) & hidden_columns
        if len(rename_targets_already_in_columns) > 0:
            raise ValueError(
                f"rename targets are already in column names: {rename_targets_already_in_columns}. Did you "
                f"miss to set remove_columns=True in a previous call of cast_document_type?"
            )

        new_hf_dataset = new_hf_dataset.rename_columns(column_mapping=field_mapping)

        new_dataset = IterableDataset.from_hf_dataset(
            new_hf_dataset,
            hidden_columns=hidden_columns,
            document_type=new_document_type,
            document_converters=self.document_converters,
        )

        return new_dataset

    def take(self, n) -> "IterableDataset":
        return self.apply_hf_func(datasets.IterableDataset.take, n=n)


def get_pie_dataset_type(
    hf_dataset: Union[datasets.Dataset, datasets.IterableDataset]
) -> Union[Type[Dataset], Type[IterableDataset]]:
    if isinstance(hf_dataset, datasets.Dataset):
        return Dataset
    elif isinstance(hf_dataset, datasets.IterableDataset):
        return IterableDataset
    else:
        raise TypeError(
            f"the dataset must be of type Dataset or IterableDataset, but is of type {type(hf_dataset)}"
        )


def _add_dset_name_to_document(doc: Document, name: str, clear_metadata: bool) -> Document:
    if not hasattr(doc, "metadata"):
        raise ValueError(
            f"Document does not have metadata attribute which required to save the dataset name: {doc}"
        )
    # Keep the old name if available
    if "dataset_name" in doc.metadata:
        name = doc.metadata["dataset_name"]
    if clear_metadata:
        doc.metadata = {}
    doc.metadata["dataset_name"] = name
    return doc


def concatenate_datasets(
    dsets: Union[
        List[Dataset], List[IterableDataset], Dict[str, Dataset], Dict[str, IterableDataset]
    ],
    clear_metadata: bool,
) -> Union[Dataset, IterableDataset]:
    """Concatenate multiple datasets into a single dataset. The datasets must have the same
    document type. Dataset name will be saved in Metadata.

    Args:
        dsets: A list of datasets or a dictionary with dataset names as keys and datasets as values. If
            a dictionary is provided, the dataset names will be added to the documents as metadata.
        clear_metadata: Whether to clear the metadata before concatenating.
    Returns:
        A new dataset that is the concatenation of the input datasets.
    """

    if isinstance(dsets, dict):
        dsets = [
            dset.map(
                _add_dset_name_to_document,
                fn_kwargs={"name": name, "clear_metadata": clear_metadata},
            )
            for name, dset in dsets.items()
        ]

    if len(dsets) == 0:
        raise ValueError("No datasets to concatenate")

    document_type = dsets[0].document_type
    for doc in dsets[1:]:
        if not doc.document_type == document_type:
            raise ValueError("All datasets must have the same document type to concatenate")

    result_hf = datasets.concatenate_datasets(dsets)
    pie_dataset_type = get_pie_dataset_type(dsets[0])

    return pie_dataset_type.from_hf_dataset(result_hf, document_type=document_type)



from .builder import ArrowBasedBuilder, GeneratorBasedBuilder
from .dataset import Dataset, IterableDataset, concatenate_datasets
from .dataset_dict import DatasetDict, concatenate_dataset_dicts, load_dataset

__all__ = [
    "GeneratorBasedBuilder",
    "ArrowBasedBuilder",
    "Dataset",
    "IterableDataset",
    "DatasetDict",
    "load_dataset",
    "concatenate_datasets",
    "concatenate_dataset_dicts",
]



import abc
from typing import Any, Callable, Dict, Optional, Type, Union, overload

import datasets
from pie_core import Document
from pie_core.utils.hydra import resolve_target

from .dataset import (
    Dataset,
    DocumentConvertersType,
    IterableDataset,
    decorate_convert_document_back,
    get_pie_dataset_type,
)


def get_general_dataset_builder_parent_class(
    obj: datasets.builder.DatasetBuilder,
) -> Type[datasets.builder.DatasetBuilder]:
    general_dataset_builder_parent_classes = [
        cls
        for cls in datasets.builder.DatasetBuilder.__subclasses__()
        if cls != PieDatasetBuilder and isinstance(obj, cls)
    ]
    if len(general_dataset_builder_parent_classes) != 1:
        raise TypeError("can not determine general dataset builder parent class of the object")
    return general_dataset_builder_parent_classes[0]


class PieDatasetBuilder(datasets.builder.DatasetBuilder):
    # The default pytorch-ie document type for the dataset.
    DOCUMENT_TYPE: Optional[Type[Document]] = None
    # A mapping from config names to PIE document types. Use this to specify individual
    # document types per config.
    DOCUMENT_TYPES: Dict[str, Type[Document]] = {}

    # The default path to the Huggingface dataset loading script that will be used as base dataset.
    BASE_DATASET_PATH: Optional[str] = None
    # A mapping from config names to Huggingface dataset loading script paths. Use this to specify individual
    # base datasets for each config.
    BASE_DATASET_PATHS: Dict[str, str] = {}

    # The default revision (e.g. git commit) of the Huggingface dataset loading script that will be used
    # as base dataset.
    BASE_DATASET_REVISION: Optional[str] = None
    # A mapping from config names to revisions (e.g. git commits) of the Huggingface dataset loading script
    # that will be used as base dataset.
    BASE_DATASET_REVISIONS: Dict[str, str] = {}

    # Define kwargs to create base configs. This should contain config names as keys
    # and the respective config kwargs dicts as values. If the config name is not contained, a new entry
    # {"name": config_name} will be created for it, i.e. the config name is passed as base config name.
    # This default behaviour can be disabled by setting BASE_CONFIG_KWARGS_DICT to None.
    BASE_CONFIG_KWARGS_DICT: Optional[Dict[Optional[str], Dict[str, Any]]] = {}
    # Define base builder kwargs. This should contain config names as keys and the respective
    # builder kwargs dicts as values.
    BASE_BUILDER_KWARGS_DICT: Optional[Dict[Optional[str], Dict[str, Any]]] = None

    # Define document converters. This should be a mapping from document types as keys to the respective
    # document converters as values. The document converters can be either callables or dicts
    # that map from original field names to new field names. If a callable is provided, it will be used to
    # convert the document. If a dict is provided, it will be used to rename the fields of the
    # document (this is done by renaming the columns which is much more efficient).
    DOCUMENT_CONVERTERS: DocumentConvertersType = {}

    def __init__(
        self,
        base_dataset_kwargs: Optional[Dict[str, Any]] = None,
        document_converters: Optional[
            Dict[Union[Type[Document], str], Union[Callable[..., Document], Dict[str, str], str]]
        ] = None,
        **kwargs,
    ):
        self.base_builder = None
        config_name = kwargs.get("config_name", None)
        base_dataset_path = self.BASE_DATASET_PATHS.get(config_name, self.BASE_DATASET_PATH)
        if base_dataset_path is not None:
            base_dataset_kwargs = base_dataset_kwargs or {}
            base_builder_kwargs: Dict[str, Any] = {}

            # get base config kwargs from mapping
            if self.BASE_CONFIG_KWARGS_DICT is not None:
                if config_name in self.BASE_CONFIG_KWARGS_DICT:
                    config_kwargs = self.BASE_CONFIG_KWARGS_DICT[config_name]
                else:
                    # if the config name is not in BASE_CONFIG_KWARGS_DICT,
                    # we pass it as base config name
                    config_kwargs = {"name": config_name}
                base_builder_kwargs.update(config_kwargs)

            # get base builder kwargs from mapping
            if self.BASE_BUILDER_KWARGS_DICT is not None:
                base_builder_kwargs.update(self.BASE_BUILDER_KWARGS_DICT[config_name])

            revision = self.BASE_DATASET_REVISIONS.get(config_name, self.BASE_DATASET_REVISION)
            if revision is not None:
                base_builder_kwargs["revision"] = revision

            base_builder_kwargs.update(base_dataset_kwargs)
            self.base_builder = datasets.load.load_dataset_builder(
                path=base_dataset_path,
                **base_builder_kwargs,
            )
            # Ensure that self and self.base_builder are derived from the same subclass of
            # datasets.builder.DatasetBuilder.
            base_builder_general_parent_class = get_general_dataset_builder_parent_class(
                self.base_builder
            )
            self_general_parent_class = get_general_dataset_builder_parent_class(self)
            if base_builder_general_parent_class != self_general_parent_class:
                raise TypeError(
                    f"The PyTorch-IE dataset builder class '{type(self).__name__}' is derived from "
                    f"{self_general_parent_class}, but the base builder is not which is not allowed. The base builder "
                    f"is of type '{type(self.base_builder).__name__}' that is derived from "
                    f"{base_builder_general_parent_class}. Consider to derive your PyTorch-IE dataset builder "
                    f"'{type(self).__name__}' from a PyTorch-IE variant of "
                    f"'{base_builder_general_parent_class.__name__}'."
                )

            # append the base_builder config_id to the hash, otherwise the base_builder config arguments
            # are not respected in the cache fingerprint
            if "hash" in kwargs:
                kwargs["hash"] = f"{kwargs['hash']}-{self.base_builder.config_id}"

            # set base path to base builder base path. This is required so that the download manager
            # works correctly with relative paths.
            kwargs["base_path"] = self.base_builder.base_path

        super().__init__(**kwargs)

        self._document_converters = dict(self.DOCUMENT_CONVERTERS)
        if document_converters is not None:
            for document_type_or_str, document_converter_or_str in document_converters.items():
                document_type = resolve_target(document_type_or_str)
                if isinstance(document_type, type) and issubclass(document_type, Document):
                    document_converter: Union[Callable[..., Any], dict[str, str]]
                    if isinstance(document_converter_or_str, str):
                        document_converter = resolve_target(document_converter_or_str)
                    else:
                        document_converter = document_converter_or_str

                    self._document_converters[document_type] = document_converter
                else:
                    raise TypeError(
                        f"The key '{document_type_or_str}' for one of the converters "
                        f"can not be resolved to a document type."
                    )

    def _info(self):
        return self.base_builder._info()

    def _split_generators(self, dl_manager):
        return self.base_builder._split_generators(dl_manager)

    @property
    def document_type(self) -> Optional[Type[Document]]:
        return self.DOCUMENT_TYPES.get(self.config.name, self.DOCUMENT_TYPE)

    @property
    def document_converters(self) -> DocumentConvertersType:
        return self._document_converters

    @abc.abstractmethod
    def _generate_document(self, example, **kwargs):
        pass

    def _generate_document_kwargs(self, dataset):
        return None

    def _generate_example(self, document: Document, **kwargs) -> Dict[str, Any]:
        raise NotImplementedError

    def _generate_example_kwargs(
        self, dataset: Union[Dataset, IterableDataset]
    ) -> Optional[Dict[str, Any]]:
        return None  # pragma: no cover

    @overload  # type: ignore
    def _convert_dataset_single(self, dataset: datasets.IterableDataset) -> IterableDataset: ...

    @overload  # type: ignore
    def _convert_dataset_single(self, dataset: datasets.Dataset) -> Dataset: ...

    def _convert_dataset_single(
        self, dataset: Union[datasets.Dataset, datasets.IterableDataset]
    ) -> Union[Dataset, IterableDataset]:
        document_type = self.document_type
        if document_type is None:
            raise TypeError(
                f"the builder has no DOCUMENT_TYPE or DOCUMENT_TYPES[{self.config.name}] defined"
            )

        fn = decorate_convert_document_back(self._generate_document)
        fn_kwargs = self._generate_document_kwargs(dataset)
        mapped_dataset = dataset.map(fn, fn_kwargs=fn_kwargs)
        dataset_type = get_pie_dataset_type(mapped_dataset)
        result = dataset_type.from_hf_dataset(
            dataset=mapped_dataset,
            document_type=document_type,
            document_converters=dict(self.document_converters),
        )
        return result

    @overload  # type: ignore
    def _convert_datasets(self, datasets: datasets.DatasetDict) -> datasets.DatasetDict: ...

    @overload  # type: ignore
    def _convert_datasets(
        self, datasets: datasets.IterableDatasetDict
    ) -> datasets.IterableDatasetDict: ...

    @overload  # type: ignore
    def _convert_datasets(self, datasets: datasets.IterableDataset) -> IterableDataset: ...

    @overload  # type: ignore
    def _convert_datasets(self, datasets: datasets.Dataset) -> Dataset: ...

    def _convert_datasets(
        self,
        datasets: Union[
            datasets.Dataset,
            datasets.IterableDataset,
            datasets.DatasetDict,
            datasets.IterableDatasetDict,
        ],
    ) -> Union[Dataset, IterableDataset, datasets.DatasetDict, datasets.IterableDatasetDict]:
        if isinstance(datasets, dict):
            return type(datasets)(
                {k: self._convert_dataset_single(v) for k, v in datasets.items()}
            )
        else:
            return self._convert_dataset_single(datasets)

    def as_dataset(
        self,
        split: Optional[datasets.Split] = None,
        run_post_process=True,
        verification_mode: Optional[Union[datasets.VerificationMode, str]] = None,
        ignore_verifications="deprecated",
        in_memory=False,
    ) -> Union[Dataset, datasets.DatasetDict]:
        dataset = super().as_dataset(
            split=split,
            run_post_process=run_post_process,
            ignore_verifications=ignore_verifications,
            in_memory=in_memory,
            verification_mode=verification_mode,
        )
        converted_datasets = self._convert_datasets(datasets=dataset)
        return converted_datasets

    def as_streaming_dataset(
        self,
        split: Optional[str] = None,
        base_path: Optional[str] = None,
    ) -> Union[IterableDataset, datasets.IterableDatasetDict]:  # type: ignore
        dataset: Union[
            datasets.IterableDataset, datasets.IterableDatasetDict
        ] = super().as_streaming_dataset(
            split=split, base_path=base_path
        )  # type: ignore
        converted_datasets = self._convert_datasets(datasets=dataset)
        return converted_datasets


class GeneratorBasedBuilder(PieDatasetBuilder, datasets.builder.GeneratorBasedBuilder):
    def _generate_examples(self, *args, **kwargs):
        return self.base_builder._generate_examples(*args, **kwargs)


class ArrowBasedBuilder(PieDatasetBuilder, datasets.builder.ArrowBasedBuilder):
    def _generate_tables(self, *args, **kwargs):
        return self.base_builder._generate_tables(*args, **kwargs)
